import asyncio as aio
import json
import logging
import typing as ty

import aio_mqtt
from bleak import BleakError

from .devices.base import Device

# linux only but don't crash on other systems
try:
    from txdbus.error import RemoteError  # noqa
except ImportError:
    RemoteError = BleakError


logger = logging.getLogger(__name__)

CONFIG_MQTT_NAMESPACE = 'homeassistant'
SENSOR_STATE_TOPIC = 'state'
BLUETOOTH_ERROR_RECONNECTION_TIMEOUT = 60
FAILURE_LIMIT = 5


ListOfConnectionErrors = (
    ConnectionError,
    BleakError,
    RemoteError,
    aio.TimeoutError,
    # AttributeError: 'NoneType' object has no attribute 'callRemote'
    AttributeError,
)


def hardware_exception_occured(exception):
    ex_str = str(exception)
    return (
        'org.freedesktop.DBus.Error.ServiceUnknown' in ex_str or
        'org.freedesktop.DBus.Error.NoReply' in ex_str or
        'org.bluez.Error.NotReady' in ex_str
    )


ListOfMQTTConnectionErrors = (
        aio_mqtt.ConnectionLostError,
        aio_mqtt.ConnectionClosedError,
        aio_mqtt.ServerDiedError,
)


class Ble2Mqtt:
    TOPIC_ROOT = 'ble2mqtt'
    BRIDGE_TOPIC = 'bridge'

    def __init__(
            self,
            host: str,
            port: int = None,
            user: ty.Optional[str] = None,
            password: ty.Optional[str] = None,
            reconnection_interval: int = 10,
            loop: ty.Optional[aio.AbstractEventLoop] = None,
    ) -> None:
        self._mqtt_host = host
        self._mqtt_port = port
        self._mqtt_user = user
        self._mqtt_password = password

        self._reconnection_interval = reconnection_interval
        self._loop = loop or aio.get_event_loop()
        self._client = aio_mqtt.Client(loop=self._loop)
        self._root_tasks = []

        self._device_manage_tasks = {}

        self.availability_topic = '/'.join((
            self.TOPIC_ROOT,
            self.BRIDGE_TOPIC,
            SENSOR_STATE_TOPIC,
        ))

        self.device_registry: ty.List[Device] = []
        self.bluetooth_restarting = False

    async def start(self):
        finished, unfinished = await aio.wait(
            [
                self._loop.create_task(self._connect_forever()),
                self._loop.create_task(self._handle_messages()),
            ],
            return_when=aio.FIRST_COMPLETED,
        )
        for t in unfinished:
            t.cancel()
        await aio.wait(unfinished)
        for t in finished:
            # forward exception from task if any
            t.result()

    @staticmethod
    async def stop_task(task):
        logger.debug(f'stop_task {task=}')
        task.cancel()
        try:
            await task
        except aio.CancelledError:
            logger.debug(f'{task} is now cancelled')

    async def close(self) -> None:
        for task in self._root_tasks:
            await self.stop_task(task)
        for k, task in self._device_manage_tasks.items():
            await self.stop_task(task)
        if self._client.is_connected():
            await self._client.disconnect()

    def _get_topic(self, dev_id, subtopic, *args):
        return '/'.join((self.TOPIC_ROOT, dev_id, subtopic, *args))

    def register(self, device: Device):
        if not device:
            return
        self.device_registry.append(device)

    @property
    def subscribed_topics(self):
        return [
            '/'.join((self.TOPIC_ROOT, topic))
            for device in self.device_registry
            for topic in device.subscribed_topics
        ]

    async def publish_topic_callback(self, topic, value):
        logger.debug(f'call publish callback {topic=} {value=}')
        await self._client.publish(
            aio_mqtt.PublishableMessage(
                topic_name='/'.join((self.TOPIC_ROOT, topic)),
                payload=value,
                qos=aio_mqtt.QOSLevel.QOS_1,
            ),
        )

    async def _handle_messages(self) -> None:
        async for message in self._client.delivered_messages(
            f'{self.TOPIC_ROOT}/#',
        ):
            logger.debug(message)
            while True:
                if message.topic_name not in self.subscribed_topics:
                    continue

                topic_wo_prefix = message.topic_name.removeprefix(
                    f'{self.TOPIC_ROOT}/',
                )
                for _device in self.device_registry:
                    if topic_wo_prefix in _device.subscribed_topics:
                        device = _device
                        break
                else:
                    raise NotImplementedError('Unknown topic')
                if not await device.client.is_connected():
                    logger.warning(
                        f'Received topic {topic_wo_prefix} '
                        f'with {message.payload} '
                        f' but {device.client} is offline',
                    )
                    continue

                try:
                    value = json.loads(message.payload)
                except ValueError:
                    value = message.payload.decode()

                await device.add_incoming_message(topic_wo_prefix, value)
                break

            await aio.sleep(1)

    async def send_device_config(self, device: Device):
        device_info = {
            'identifiers': [
                device.unique_id,
            ],
            'name': device.unique_id,
            'sw_version': device.version,
            'model': device.model,
            'manufacturer': device.manufacturer,
        }

        def get_generic_vals(entity: dict):
            name = entity.pop('name')
            result = {
                'name': f'{name}_{device.dev_id}',
                'unique_id': f'{name}_{device.dev_id}',
                'device': device_info,
            }
            icon = entity.pop('icon', None)
            if icon:
                result['icon'] = f'mdi:{icon}'
            entity.pop('topic', None)
            entity.pop('json', None)
            entity.pop('main_value', None)
            result.update(entity)
            return result

        messages_to_send = []
        for cls, entities in device.entities.items():
            if cls == 'switch':
                for entity in entities:
                    entity_name = entity['name']
                    state_topic = self._get_topic(device.unique_id, entity_name)
                    command_topic = '/'.join((state_topic, device.SET_POSTFIX))
                    config_topic = '/'.join((
                        CONFIG_MQTT_NAMESPACE,
                        cls,
                        device.dev_id,
                        entity_name,
                        'config',
                    ))
                    payload = json.dumps({
                        **get_generic_vals(entity),
                        'state_topic': state_topic,
                        'command_topic': command_topic,
                    })
                    logger.debug(
                        f'Publish config {config_topic=}: {payload=}',
                    )
                    messages_to_send.append(
                        aio_mqtt.PublishableMessage(
                            topic_name=config_topic,
                            payload=payload,
                            qos=aio_mqtt.QOSLevel.QOS_1,
                            retain=True,
                        ),
                    )
                    # TODO: send real state on receiving status from a device
                    logger.debug(f'Publish initial state {state_topic=}')
                    await self._client.publish(
                        aio_mqtt.PublishableMessage(
                            topic_name=state_topic,
                            payload='OFF',
                            qos=aio_mqtt.QOSLevel.QOS_1,
                        ),
                    )
            if cls == 'sensor':
                for entity in entities:
                    entity_name = entity['name']
                    state_topic = self._get_topic(
                        device.unique_id,
                        entity.get('topic', SENSOR_STATE_TOPIC),
                    )
                    config_topic = '/'.join((
                        CONFIG_MQTT_NAMESPACE,
                        cls,
                        device.dev_id,
                        entity_name,
                        'config',
                    ))
                    if entity.get('json') and entity.get('main_value'):
                        state_topic_part = {
                            'json_attributes_topic': state_topic,
                            'state_topic': state_topic,
                            'value_template':
                                f'{{{{ value_json.{entity["main_value"]} }}}}',
                        }
                    else:
                        state_topic_part = {
                            'state_topic': state_topic,
                            'value_template':
                                f'{{{{ value_json.{entity_name} }}}}',
                        }

                    payload = json.dumps({
                        **get_generic_vals(entity),
                        **state_topic_part,
                    })
                    logger.debug(
                        f'Publish config {config_topic=}: {payload=}',
                    )
                    messages_to_send.append(
                        aio_mqtt.PublishableMessage(
                            topic_name=config_topic,
                            payload=payload,
                            qos=aio_mqtt.QOSLevel.QOS_1,
                            retain=True,
                        ),
                    )
            if cls == 'light':
                for entity in entities:
                    entity_name = entity['name']
                    state_topic = self._get_topic(device.unique_id, entity_name)
                    set_topic = self._get_topic(
                        device.unique_id,
                        entity_name,
                        device.SET_POSTFIX,
                    )
                    config_topic = '/'.join((
                        CONFIG_MQTT_NAMESPACE,
                        cls,
                        device.dev_id,
                        entity_name,
                        'config',
                    ))
                    payload = json.dumps({
                        **get_generic_vals(entity),
                        'schema': 'json',
                        'rgb': entity.get('rgb', True),
                        'brightness': entity.get('brightness', True),
                        'state_topic': state_topic,
                        'command_topic': set_topic,
                    })
                    logger.debug(
                        f'Publish config {config_topic=}: {payload=}',
                    )
                    messages_to_send.append(
                        aio_mqtt.PublishableMessage(
                            topic_name=config_topic,
                            payload=payload,
                            qos=aio_mqtt.QOSLevel.QOS_1,
                            retain=True,
                        ),
                    )
        await aio.gather(*[
            self._client.publish(message)
            for message in messages_to_send
        ])

    async def restart_bluetooth(self):
        # EXPERIMENTAL: restart bluetooth on errors
        if self.bluetooth_restarting:
            await aio.sleep(7)
            return
        self.bluetooth_restarting = True
        logger.warning('Restarting bluetoothd...')
        proc = await aio.create_subprocess_exec(
            'hciconfig', 'hci0', 'down',
        )
        await proc.wait()
        proc = await aio.create_subprocess_exec(
            '/etc/init.d/bluetoothd', 'restart',
        )
        await proc.wait()
        await aio.sleep(3)
        proc = await aio.create_subprocess_exec(
            'hciconfig', 'hci0', 'up',
        )
        await proc.wait()
        await aio.sleep(3)
        logger.warning('Restarting bluetoothd finished')
        self.bluetooth_restarting = False

    async def manage_device(self, device: Device):
        logger.debug(f'Start managing {device=}')
        failure_count = 0
        while True:
            logger.debug(f'Connecting to {device=}')
            connect_task = self._loop.create_task(device.connect())
            finished, unfinished = await aio.wait(
                [connect_task],
                timeout=device.CONNECTION_TIMEOUT,
            )
            try:
                if connect_task not in finished:
                    logger.debug(f'Stop task {connect_task=} {device=}')
                    await self.stop_task(connect_task)
                    raise ConnectionError(f'Task is timed out {device=}')
                else:
                    t, = finished
                    disconnect_fut = t.result()
            except ListOfConnectionErrors as e:
                logger.warning(f'Error while connecting to {device=}, {e}')
                await device.close()
                if 'DBus.Error.LimitsExceeded' in str(e):
                    raise
                if hardware_exception_occured(e):
                    await self.restart_bluetooth()
                    await aio.sleep(3)
                    continue
                elif 'org.bluez.Error.' in str(e) or \
                        'org.freedesktop.DBus.Error.' in str(e):
                    failure_count += 1
                    if failure_count >= FAILURE_LIMIT:
                        await self.restart_bluetooth()
                        # raise
                    logger.error(
                        f'Sleep for {BLUETOOTH_ERROR_RECONNECTION_TIMEOUT} '
                        f'secs due to error in bluetooth, '
                        f'{device=}, exception={e}',
                    )
                    await aio.sleep(BLUETOOTH_ERROR_RECONNECTION_TIMEOUT)
                continue

            try:
                # retrieve version and details
                logger.debug(f'get_device_data {device=}')
                await device.get_device_data()
                failure_count = 0
            except ListOfConnectionErrors:
                logger.exception(f'Cannot get initial info {device=}')
                await device.close()
                continue

            try:
                await self.send_device_config(device)

                if device.subscribed_topics:
                    await self._client.subscribe(*[
                        (
                            '/'.join((self.TOPIC_ROOT, topic)),
                            aio_mqtt.QOSLevel.QOS_1,
                        )
                        for topic in device.subscribed_topics
                    ])
            except aio_mqtt.Error:
                logger.exception(f'Cannot subscribe to topics {device=}')
                await device.client.disconnect()
                return

            try:
                logger.info(
                    f'Start device {device=} handle task and wait '
                    f'for disconnect',
                )
                finished, unfinished = await aio.wait(
                    [
                        disconnect_fut,
                        self._loop.create_task(
                            device.handle(self.publish_topic_callback),
                        ),
                        self._loop.create_task(
                            device.handle_messages(self.publish_topic_callback),
                        ),
                    ],
                    return_when=aio.FIRST_COMPLETED,
                )
                logger.debug(
                    f'Handle tasks finished. {device=} disconnected. '
                    f'{finished=} {unfinished}',
                )
                for t in finished:
                    logger.debug(f'Fetching result {device=} {t=}')
                    t.result()
                for t in unfinished:
                    t.cancel()
                logger.debug(f'wait for cancelling tasks for {device=}')
                await aio.wait(unfinished)
            except ListOfMQTTConnectionErrors:
                logger.exception('Stop manage task on MQTT connection error')
                await device.close()
                return
            except Exception as e:
                if hardware_exception_occured(e):
                    await self.restart_bluetooth()
                logger.exception(f'{device=} raised an error')
            finally:
                logger.debug(f'unsubscribe from topics for {device=}')
                try:
                    if device.subscribed_topics:
                        await self._client.unsubscribe(*[
                            '/'.join((self.TOPIC_ROOT, topic))
                            for topic in device.subscribed_topics
                        ])
                except aio_mqtt.ConnectionClosedError:
                    logger.exception(
                        'Stop manage task on MQTT connection error',
                    )
                    return

                except Exception as e:
                    logger.exception(
                        f'Couldn\'t stop all tasks for {device=} {e}',
                    )

                try:
                    await device.close()
                except ListOfConnectionErrors as e:
                    # if close raises an error it might relate to
                    # temporary down of bluetooth device
                    # if so, sleep for a min and try to reconnect
                    failure_count += 1
                    if failure_count >= FAILURE_LIMIT:
                        await self.restart_bluetooth()

                    # 'org.freedesktop.DBus.Error.ServiceUnknown'
                    # 'org.freedesktop.DBus.Error.NoReply'
                    elif 'org.freedesktop.DBus.Error.' in str(e):
                        device.client = None
                        logger.error(
                            f'Sleep for {BLUETOOTH_ERROR_RECONNECTION_TIMEOUT} '
                            f'secs due to error in bluetooth, '
                            f'{device=}, exception={e}',
                        )
                        await aio.sleep(BLUETOOTH_ERROR_RECONNECTION_TIMEOUT)

            logger.info(
                f'Sleep for {device.RECONNECTION_TIMEOUT} secs to '
                f'reconnect to {device=}',
            )
            await aio.sleep(device.RECONNECTION_TIMEOUT)

    async def create_device_manage_tasks(self):
        tasks = []
        for dev in self.device_registry:
            task = self._loop.create_task(self.manage_device(dev))
            self._device_manage_tasks[dev] = task
            tasks.append(task)
        return tasks

    async def stop_device_manage_tasks(self):
        for dev in list(self._device_manage_tasks.keys()):
            logger.info(f'Stopping manage task for {dev=}')
            task = self._device_manage_tasks.pop(dev)
            task.cancel()
            await aio.wait([task])

    async def _run_device_tasks(self, mqtt_connection_fut: aio.Future) -> None:
        tasks = await self.create_device_manage_tasks()
        logger.debug("Wait for network interruptions...")
        finished, unfinished = await aio.wait(
            [
                mqtt_connection_fut,
                *tasks,
            ],
            return_when=aio.FIRST_COMPLETED,
        )
        for t in finished:
            try:
                t.result()
            except Exception:
                logger.exception('Root task has raised an exception')
        for t in unfinished:
            t.cancel()
        try:
            await aio.wait(unfinished)
        except aio.CancelledError:
            pass

    async def _connect_forever(self) -> None:
        while True:
            try:
                connect_result = await self._client.connect(
                    host=self._mqtt_host,
                    port=self._mqtt_port,
                    username=self._mqtt_user,
                    password=self._mqtt_password,
                    will_message=aio_mqtt.PublishableMessage(
                        topic_name=self.availability_topic,
                        payload='offline',
                        qos=aio_mqtt.QOSLevel.QOS_1,
                        retain=True,
                    ),
                )
                logger.info(f'Connected to {self._mqtt_host}')
                await self._client.publish(
                    aio_mqtt.PublishableMessage(
                        topic_name=self.availability_topic,
                        payload='online',
                        qos=aio_mqtt.QOSLevel.QOS_1,
                        retain=True,
                    ),
                )
                await self._run_device_tasks(connect_result.disconnect_reason)

            except aio.CancelledError:
                raise

            except aio_mqtt.AccessRefusedError as e:
                await self.stop_device_manage_tasks()
                logger.error("Access refused", exc_info=e)

            except ListOfMQTTConnectionErrors as e:
                try:
                    await self.stop_device_manage_tasks()
                except Exception as e1:
                    logger.exception(e1)
                logger.error(
                    "Connection lost. Will retry in %d seconds",
                    self._reconnection_interval,
                    exc_info=e,
                )
                await aio.sleep(self._reconnection_interval)

            except aio_mqtt.ConnectionCloseForcedError as e:
                logger.error("Connection close forced", exc_info=e)
                return

            except Exception as e:
                logger.error(
                    "Unhandled exception during connecting",
                    exc_info=e,
                )
                try:
                    await self._client.publish(
                        aio_mqtt.PublishableMessage(
                            topic_name=self.availability_topic,
                            payload='offline',
                            qos=aio_mqtt.QOSLevel.QOS_1,
                            retain=True,
                        ),
                    )
                except Exception:
                    pass
                return
            else:
                try:
                    await self._client.publish(
                        aio_mqtt.PublishableMessage(
                            topic_name=self.availability_topic,
                            payload='offline',
                            qos=aio_mqtt.QOSLevel.QOS_1,
                            retain=True,
                        ),
                    )
                except Exception:
                    pass
                logger.info("Disconnected")
                return



import asyncio as aio
import logging
import typing as ty

from bleak import BleakClient, BleakError

logger = logging.getLogger(__name__)
registered_device_types = {}


class RegisteredType(type):
    def __new__(mcs, clsname, superclasses, attributedict):
        newclass = type.__new__(mcs, clsname, superclasses, attributedict)
        # condition to prevent base class registration
        if superclasses:
            if newclass.NAME is not None:
                registered_device_types[newclass.NAME] = newclass
        return newclass


class BaseDevice(metaclass=RegisteredType):
    NAME = None

    def __init__(self, *args, loop, **kwargs):
        self._loop = loop
        self.client: BleakClient = None
        self.bt_lock = aio.Lock()

    async def close(self):
        pass

    async def _read_with_timeout(self, char, timeout=5):
        try:
            result = await aio.wait_for(
                self.client.read_gatt_char(char),
                timeout=timeout,
                loop=self._loop,
            )
        except Exception:
            logger.exception('Cannot connect to device')
            result = None
        return result

    @staticmethod
    def transform_value(value):
        if not isinstance(value, str):
            return value
        vl = value.lower()
        if vl in ['0', 'off', 'no']:
            return 'OFF'
        elif vl in ['1', 'on', 'yes']:
            return 'ON'
        return value

    @property
    def unique_id(self):
        raise None


class Device(BaseDevice):
    MQTT_VALUES = None
    ON_OFF = False
    SET_POSTFIX = 'set'
    RECONNECTION_TIMEOUT = 3
    REQUIRE_CONNECTION = False
    MAC_TYPE = 'public'
    MANUFACTURER = None
    CONNECTION_TIMEOUT = 60

    def __init__(self, mac, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.disconnected_future: ty.Optional[aio.Future] = None
        self.message_queue = aio.Queue()
        self._mac = mac
        self._model = None
        self._version = None
        self._manufacturer = self.MANUFACTURER
        self.connection_event = aio.Event()

    def get_entity_from_topic(self, topic: str):
        return topic.removesuffix(self.SET_POSTFIX).removeprefix(
            self.unique_id,
        ).strip('/')

    @property
    def subscribed_topics(self):
        return [
            f'{self.unique_id}/{entity["name"]}/{self.SET_POSTFIX}'
            for cls, items in self.entities.items()
            for entity in items
            if cls in ['switch', 'light']
        ]

    @property
    def manufacturer(self):
        return self._manufacturer

    @property
    def model(self):
        return self._model

    @property
    def version(self):
        return self._version

    @property
    def dev_id(self):
        return self._mac.replace(':', '').lower()

    @property
    def unique_id(self):
        parts = [self.manufacturer, self.model, self.dev_id]
        return '_'.join([p for p in parts if p])

    @property
    def entities(self):
        return {}

    async def handle(self, *args, **kwargs):
        raise NotImplementedError()

    async def handle_messages(self, *args, **kwargs):
        while True:
            await aio.sleep(1)

    def __str__(self):
        return self.unique_id

    def __repr__(self):
        return f'<Device:{str(self)}>'

    async def add_incoming_message(self, topic: str, value):
        await self.message_queue.put({
            'topic': topic,
            'value': value,
        })

    async def get_device_data(self):
        """Here put the initial configuration for the device"""
        pass

    async def get_client(self) -> BleakClient:
        assert self.MAC_TYPE in ('public', 'random')
        return BleakClient(self._mac, address_type=self.MAC_TYPE)

    async def connect(self):
        self.client = await self.get_client()
        self.disconnected_future = self._loop.create_future()
        try:
            async with self.bt_lock:
                await self.client.connect()
                self.connection_event.set()
            self.client.set_disconnected_callback(self.on_disconnect)
        except BleakError:
            self.connection_event.clear()
            self.client.set_disconnected_callback(None)
            raise
        logger.info(f'Connected to {self.client.address}')
        return self.disconnected_future

    def on_disconnect(self, client, *args):
        logger.info(f'Client {client.address} disconnected, device={self}')
        self.connection_event.clear()
        if self.disconnected_future.done() or \
                self.disconnected_future.cancelled():
            raise NotImplementedError(
                f'disconnected_future for device={self} is '
                f'{self.disconnected_future}',
            )
        self.disconnected_future.set_result(client.address)
        self.client.set_disconnected_callback(None)
        self.client = None

    async def close(self):
        try:
            if self.client and await self.client.is_connected():
                await self.client.disconnect()
        # exception on macos when checking for is_connected()
        except AttributeError:
            pass



import logging
import uuid
from dataclasses import dataclass

from .base import Device
from .xiaomi_base import XiaomiHumidityTemperature

logger = logging.getLogger(__name__)

MJHT_DATA = uuid.UUID('226caa55-6476-4566-7562-66734470666d')
MJHT_BATTERY = uuid.UUID('00002a19-0000-1000-8000-00805f9b34fb')


@dataclass
class SensorState:
    battery: int
    temperature: float
    humidity: float

    @classmethod
    def from_data(cls, sensor_data, battery_data):
        # b'T=23.6 H=39.6\x00'
        t, h = tuple(
            float(x.split('=')[1])
            for x in sensor_data.decode().strip('\0').split(' ')
        )
        return cls(
            temperature=t,
            humidity=h,
            battery=int(ord(battery_data)),
        )


class XiaomiHumidityTemperatureV1(XiaomiHumidityTemperature, Device):
    NAME = 'xiaomihtv1'
    REQUIRE_CONNECTION = True
    DATA_CHAR = MJHT_DATA
    BATTERY_CHAR = MJHT_BATTERY
    SENSOR_CLASS = SensorState

    def filter_notifications(self, sender):
        # sender is 0xd or several requests it becomes
        # /org/bluez/hci0/dev_58_2D_34_32_E0_69/service000c/char000d
        return (
            sender == 0xd or
            isinstance(sender, str) and sender.endswith('000d')
        )



import asyncio as aio
import json
import logging
import os

from ble2mqtt.ble2mqtt import Ble2Mqtt

from .devices import registered_device_types


def main():
    logging.basicConfig(level='INFO')
    loop = aio.get_event_loop()

    os.environ.setdefault('BLE2MQTT_CONFIG', '/etc/ble2mqtt.json')
    config = {}
    if os.path.exists(os.environ['BLE2MQTT_CONFIG']):
        try:
            with open(os.environ['BLE2MQTT_CONFIG'], 'r') as f:
                config = json.load(f)
        except Exception:
            pass

    config = {
        'mqtt_host': 'localhost',
        'mqtt_port': 1883,
        **config,
    }

    server = Ble2Mqtt(
        reconnection_interval=10,
        loop=loop,
        host=config['mqtt_host'],
        port=config['mqtt_port'],
        user=config.get('mqtt_user'),
        password=config.get('mqtt_password'),
    )

    devices = config.get('devices') or []
    for device in devices:
        try:
            mac = device.pop('address')
            typ = device.pop('type')
        except (ValueError, IndexError):
            continue
        klass = registered_device_types[typ]
        server.register(klass(
            mac=mac,
            loop=loop,
            **device,
        ))

    try:
        loop.run_until_complete(server.start())
    except KeyboardInterrupt:
        pass

    finally:
        loop.run_until_complete(server.close())
        loop.run_until_complete(loop.shutdown_asyncgens())
        loop.close()


if __name__ == '__main__':
    main()



import asyncio as aio
import json
import logging
import uuid

from ..protocols.redmond import (ColorTarget, Kettle200State, Mode,
                                 RedmondKettle200Protocol, RunState)
from .base import Device
from .uuids import DEVICE_NAME

logger = logging.getLogger(__name__)

UUID_NORDIC_TX = uuid.UUID("6e400002-b5a3-f393-e0a9-e50e24dcca9e")
UUID_NORDIC_RX = uuid.UUID("6e400003-b5a3-f393-e0a9-e50e24dcca9e")

BOIL_ENTITY = 'boil'
HEAT_ENTITY = 'heat'  # not implemented yet
TEMPERATURE_ENTITY = 'temperature'
LIGHT_ENTITY = 'backlight'


class RedmondKettle(RedmondKettle200Protocol, Device):
    MAC_TYPE = 'random'
    NAME = 'redmond200'
    TX_CHAR = UUID_NORDIC_TX
    RX_CHAR = UUID_NORDIC_RX
    REQUIRE_CONNECTION = True
    CONNECTION_TIMEOUT = 30
    MANUFACTURER = 'Redmond'

    UPDATE_PERIOD = 5  # seconds when boiling
    STANDBY_UPDATE_PERIOD_MULTIPLIER = 12  # 12 * 5 seconds in standby mode

    def __init__(self, mac, key=b'\xff\xff\xff\xff\xff\xff\xff\xff',
                 *args, loop, **kwargs):
        super().__init__(mac, *args, loop=loop, **kwargs)
        assert isinstance(key, bytes) and len(key) == 8
        self._key = key
        self._state = None
        self._color = (255, 255, 255)
        self._brightness = 255
        self._statistics = {}

        self._update_period_multiplier = self.STANDBY_UPDATE_PERIOD_MULTIPLIER
        self.initial_status_sent = False

    @property
    def entities(self):
        return {
            'switch': [
                {
                    'name': BOIL_ENTITY,
                    'icon': 'kettle',
                },
            ],
            'sensor': [
                {
                    'name': TEMPERATURE_ENTITY,
                    'device_class': 'temperature',
                    'unit_of_measurement': '\u00b0C',
                },
                {
                    'name': 'statistics',
                    'topic': 'statistics',
                    'icon': 'chart-bar',
                    'json': True,
                    'main_value': 'number_of_starts',
                    'unit_of_measurement': ' ',
                },
            ],
            'light': [
                {
                    'name': LIGHT_ENTITY,
                },
            ],
        }

    async def get_device_data(self):
        await super().protocol_start()
        await self.login(self._key)
        model = await self._read_with_timeout(DEVICE_NAME)
        if isinstance(model, (bytes, bytearray)):
            self._model = model.decode()
        else:
            # macos can't access characteristic
            self._model = 'G200S'
        version = await self.get_version()
        if version:
            self._version = f'{version[0]}.{version[1]}'
        state = await self.get_mode()
        if state:
            self._state = state
            self.update_multiplier()
            self.initial_status_sent = False
        await self.set_time()
        await self._update_statistics()

    def update_multiplier(self, state: Kettle200State = None):
        if state is None:
            state = self._state
        self._update_period_multiplier = (
            1
            if state.state == RunState.ON and
            state.mode in [Mode.BOIL, Mode.HEAT]
            else self.STANDBY_UPDATE_PERIOD_MULTIPLIER
        )

    async def _notify_state(self, publish_topic):
        logger.info(f'[{self._mac}] send state={self._state}')
        coros = []

        state = {}
        for sensor_name, value in (
            ('temperature', self._state.temperature),
        ):
            if any(
                    x['name'] == sensor_name
                    for x in self.entities.get('sensor', [])
            ):
                state[sensor_name] = self.transform_value(value)

        if state:
            coros.append(publish_topic(
                topic='/'.join((self.unique_id, 'state')),
                value=json.dumps(state),
            ))

        # keep statistics in a separate topic
        logger.info(f'[{self._mac}] send statistics={self._statistics}')
        for sensor_name, value in (
            ('statistics', self._statistics),
        ):
            if any(
                x['name'] == sensor_name
                for x in self.entities.get('sensor', [])
            ):
                coros.append(publish_topic(
                    topic='/'.join((self.unique_id, sensor_name)),
                    value=json.dumps(value),
                ))

        lights = self.entities.get('light', [])
        for light in lights:
            if light['name'] == LIGHT_ENTITY:
                light_state = {
                    'state': (
                        RunState.ON.name
                        if self._state.state == RunState.ON and
                        self._state.mode == Mode.LIGHT
                        else RunState.OFF.name
                    ),
                    'brightness': 255,
                    'color': {
                        'r': self._color[0],
                        'g': self._color[1],
                        'b': self._color[2],
                    },
                }
                coros.append(publish_topic(
                    topic='/'.join((self.unique_id, light['name'])),
                    value=json.dumps(light_state),
                ))
        if coros:
            await aio.gather(*coros)

    async def notify_run_state(self, new_state: Kettle200State, publish_topic):
        if not self.initial_status_sent or \
                new_state.state != self._state.state or \
                new_state.mode != self._state.mode:
            state_to_str = {
                True: RunState.ON.name,
                False: RunState.OFF.name,
            }
            boil_mode = state_to_str[
                new_state.mode == Mode.BOIL and
                new_state.state == RunState.ON
            ]
            heat_mode = state_to_str[
                new_state.mode == Mode.HEAT and
                new_state.state == RunState.ON
            ]
            topics = {
                BOIL_ENTITY: boil_mode,
                HEAT_ENTITY: heat_mode,
            }
            await aio.gather(
                *[
                    publish_topic(
                        topic='/'.join((self.unique_id, topic)),
                        value=value,
                    ) for topic, value in topics.items()
                ],
                self._notify_state(publish_topic),
            )
            self.initial_status_sent = True
            self._state = new_state
            await self._notify_state(publish_topic)
        else:
            self._state = new_state
        self.update_multiplier()

    async def _update_statistics(self):
        statistics = await self.get_statistics()
        self._statistics = {
            'number_of_starts': statistics['starts'],
            'Energy spent (kWh)': round(statistics['watts_hours']/1000, 2),
            'Working time (minutes)': round(statistics['seconds_run']/60, 1),
        }

    async def handle(self, publish_topic, *args, **kwargs):
        counter = 0
        while True:
            # if boiling notify every 5 seconds, 60 sec otherwise
            new_state = await self.get_mode()
            await self.notify_run_state(new_state, publish_topic)
            counter += 1

            if counter > self.UPDATE_PERIOD * self._update_period_multiplier:
                await self._update_statistics()
                await self._notify_state(publish_topic)
                counter = 0
            await aio.sleep(1)

    async def _switch_mode(self, mode, value):
        if value == RunState.ON.name:
            try:
                if self._state.mode != mode:
                    await self.stop()
                await self.set_mode(Kettle200State(mode=mode))
            except ValueError:
                # if the MODE is the same then it returns
                # en error. Treat it as normal
                pass
            await self.run()
            next_state = RunState.ON
        else:
            await self.stop()
            next_state = RunState.OFF
        self.update_multiplier(Kettle200State(state=next_state))

    async def _switch_boil(self, value):
        await self._switch_mode(Mode.BOIL, value)

    async def _switch_backlight(self, value):
        await self._switch_mode(Mode.LIGHT, value)

    async def handle_messages(self, publish_topic, *args, **kwargs):
        while True:
            message = await self.message_queue.get()
            value = message['value']
            entity_name = self.get_entity_from_topic(message['topic'])
            if entity_name == BOIL_ENTITY:
                value = self.transform_value(value)
                logger.info(
                    f'[{self._mac}] switch kettle {entity_name=} {value=}',
                )
                while True:
                    try:
                        await self._switch_boil(value)
                        # update state to real values
                        await self.get_mode()
                        await aio.gather(
                            publish_topic(
                                topic='/'.join((self.unique_id, entity_name)),
                                value=self.transform_value(value),
                            ),
                            self._notify_state(publish_topic),
                            loop=self._loop,
                        )
                        break
                    except ConnectionError as e:
                        logger.exception(str(e))
                    await aio.sleep(5)
            if entity_name == LIGHT_ENTITY:
                logger.info(f'set backlight {value}')
                if value.get('state'):
                    await self._switch_backlight(value['state'])
                if value.get('color') or value.get('brightness'):
                    if value.get('color'):
                        color = value['color']
                        try:
                            self._color = color['r'], color['g'], color['b']
                        except ValueError:
                            return
                    if value.get('brightness'):
                        self._brightness = value['brightness']
                    await self.set_color(
                        ColorTarget.LIGHT,
                        *self._color,
                        self._brightness,
                    )



import json
import logging

from ..protocols.xiaomi import XiaomiPoller

logger = logging.getLogger(__name__)


class XiaomiHumidityTemperature(XiaomiPoller):
    SENSOR_CLASS = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._state = None

    @property
    def entities(self):
        return {
            'sensor': [
                {
                    'name': 'temperature',
                    'device_class': 'temperature',
                    'unit_of_measurement': '\u00b0C',
                },
                {
                    'name': 'humidity',
                    'device_class': 'humidity',
                    'unit_of_measurement': '%',
                },
                {
                    'name': 'battery',
                    'device_class': 'battery',
                    'unit_of_measurement': '%',
                },
            ],
        }

    async def _notify_state(self, publish_topic):
        logger.info(f'[{self}] send state={self._state}')
        state = {}
        for sensor_name, value in (
                ('temperature', self._state.temperature),
                ('humidity', self._state.humidity),
                ('battery', self._state.battery),
        ):
            if any(
                    x['name'] == sensor_name
                    for x in self.entities.get('sensor', [])
            ):
                state[sensor_name] = self.transform_value(value)

        if state:
            await publish_topic(
                topic='/'.join((self.unique_id, 'state')),
                value=json.dumps(state),
            )

    async def read_and_send_data(self, publish_topic):
        battery = await self._read_with_timeout(self.BATTERY_CHAR)
        data_bytes = await self._stack.get()
        # clear queue
        while not self._stack.empty():
            self._stack.get_nowait()
        self._state = self.SENSOR_CLASS.from_data(data_bytes, battery)
        await self._notify_state(publish_topic)



import asyncio as aio
import logging
import struct
import time
from dataclasses import dataclass
from enum import Enum

from ..devices.base import BaseDevice

logger = logging.getLogger(__name__)


BOIL_TIME_RELATIVE_DEFAULT = 0x80


class RedmondError(ValueError):
    pass


class Command(Enum):
    VERSION = 0x01
    RUN_CURRENT_MODE = 0x03
    STOP_CURRENT_MODE = 0x04
    WRITE_MODE = 0x05
    READ_MODE = 0x06
    SET_TIME = 0x6e
    WRITE_COLOR = 0x32
    READ_COLOR = 0x33
    SET_BACKLIGHT_MODE = 0x37
    GET_STATISTICS = 0x47
    GET_STARTS_COUNT = 0x50
    AUTH = 0xFF


class Mode(Enum):
    BOIL = 0x00
    HEAT = 0x01
    LIGHT = 0x03


class RunState(Enum):
    OFF = 0x00
    ON = 0x02


class ColorTarget(Enum):
    BOIL = 0x00
    LIGHT = 0x01


@dataclass
class Kettle200State:
    temperature: int = 0
    color_change_period: int = 0xf
    mode: Mode = Mode.BOIL
    target_temperature: int = 0
    sound: bool = True
    state: RunState = RunState.OFF
    boil_time: int = 0
    error: int = 0

    FORMAT = '<HH2B3H2BH'

    @classmethod
    def from_bytes(cls, response):
        # 00 00 00 00 01 16 0f 00 00 00 00 00 00 80 00 00 - wait
        # 00 00 00 00 01 14 0f 00 02 00 00 00 00 80 00 00 - boil
        # 01 00 28 00 01 19 0f 00 00 00 00 00 00 80 00 00 - 40º keep
        (
            mode,  # 0,1
            target_temp,  # 2,3
            sound,  # 4
            current_temp,  # 5
            color_change_period,  # 6-7
            state,  # 8,9
            _,  # 10,11
            _,   # 12,
            boil_time_relative,  # 13
            error,  # 14,15
        ) = struct.unpack(cls.FORMAT, response)
        return cls(
            mode=Mode(mode),
            target_temperature=target_temp,
            sound=sound,
            temperature=current_temp,
            state=RunState(state),
            boil_time=boil_time_relative - BOIL_TIME_RELATIVE_DEFAULT,
            color_change_period=color_change_period,
            error=error,
        )

    def to_bytes(self):
        return struct.pack(
            self.FORMAT,
            self.mode.value,
            self.target_temperature,
            self.sound,
            self.temperature,
            self.color_change_period,
            self.state.value,
            0,
            0,
            self.boil_time + BOIL_TIME_RELATIVE_DEFAULT,
            self.error,
        )


class RedmondKettle200Protocol(BaseDevice):
    MAGIC_START = 0x55
    MAGIC_END = 0xaa

    RX_CHAR = None
    TX_CHAR = None

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._cmd_counter = 0
        self.wait_event = aio.Event()
        self.received_data = None

    def notification_handler(self, sender: int, data: bytearray):
        logger.debug("Notification: {0}: {1}".format(
            sender,
            ' '.join(format(x, '02x') for x in data),
        ))
        if not self.wait_event.is_set():
            self.received_data = data
            self.wait_event.set()

    async def protocol_start(self):
        # we assume that every time protocol starts it uses new blank
        # BleakClient to avoid multiple char notifications on every restart
        # bug ?

        assert self.RX_CHAR and self.TX_CHAR
        # if not self.notification_started:
        assert await self.client.is_connected()
        # check for fresh client
        assert not self.client._notification_callbacks
        logger.info(f'Enable BLE notifications from [{self.client.address}]')
        async with self.bt_lock:
            try:
                await self.client.write_gatt_char(
                    self.TX_CHAR,
                    bytearray(0x01.to_bytes(2, byteorder="little")),
                    True,
                )
                await self.client.start_notify(
                    self.RX_CHAR,
                    self.notification_handler,
                )
            except Exception:
                await self.client._cleanup_all()
                raise

    async def protocol_stop(self):
        # NB: not used for now as we don't disconnect from our side
        async with self.bt_lock:
            await self.client.stop_notify(self.RX_CHAR)

    @staticmethod
    def _check_success(response,
                       error_msg="Command was not completed successfully"):
        success = response and response[0]
        if not success:
            raise RedmondError(error_msg)

    @staticmethod
    def _check_zero_response(response,
                             error_msg="Command was not completed successfully",
                             ):
        response = response and response[0]
        if response != 0:
            raise RedmondError(error_msg)

    def _get_command(self, cmd: int, payload: bytes):
        container = struct.pack(
            '<4B',
            self.MAGIC_START,
            self._cmd_counter,
            cmd,
            self.MAGIC_END,
        )
        self._cmd_counter += 1
        if self._cmd_counter > 100:
            self._cmd_counter = 0
        return bytearray(b'%b%b%b' % (container[:3], payload, container[3:]))

    async def send_command(self, cmd: Command, payload: bytes = b'',
                           wait_reply=True, timeout=15):
        command = self._get_command(cmd.value, payload)
        logger.debug(
            f'... send cmd {cmd.value:04x} ['
            f'{"".join(format(x, "02x") for x in payload)}] '
            f'{" ".join(format(x, "02x") for x in command)}',
        )
        self.wait_event.clear()
        async with self.bt_lock:
            try:
                cmd_resp = await aio.wait_for(
                    self.client.write_gatt_char(
                        self.TX_CHAR,
                        command,
                        True,
                    ),
                    timeout=timeout,
                )
            except aio.TimeoutError:
                raise ConnectionError('Cannot connect to device') from None
            if wait_reply:
                await aio.wait_for(
                    self.wait_event.wait(),
                    timeout,
                )

                # extract payload from container
                cmd_resp = bytes(self.received_data[3:-1])
                self.wait_event.clear()
                self.received_data = None
        return cmd_resp

    async def login(self, key):
        logger.debug('logging in...')
        resp = await self.send_command(Command.AUTH, key, True)
        self._check_success(resp, "Not logged in")

    async def get_version(self):
        logger.debug('fetching version...')
        resp = await self.send_command(Command.VERSION, b'', True)
        version = tuple(resp)
        logger.info(f'version: {version=}')
        return version

    async def set_time(self, ts=None):
        if ts is None:
            ts = time.time()
        ts = int(ts)
        offset = time.timezone \
            if (time.localtime().tm_isdst == 0) else time.altzone
        logger.debug(f'Setting time {ts=} {offset=}')
        resp = await self.send_command(
            Command.SET_TIME,
            struct.pack('<ii', ts, -offset * 60 * 60),
        )
        self._check_zero_response(resp, 'Cannot set time')

    async def get_mode(self):
        logger.debug('Get mode...')
        response = await self.send_command(Command.READ_MODE)
        return Kettle200State.from_bytes(response)

    async def set_mode(self, state: Kettle200State):
        logger.debug('Set mode...')
        resp = await self.send_command(
            Command.WRITE_MODE,
            state.to_bytes(),
        )
        success = resp[0]
        if not success:
            raise RedmondError('Cannot set mode')

    async def run(self):
        logger.debug('Run mode')
        resp = await self.send_command(Command.RUN_CURRENT_MODE)
        self._check_success(resp)

    async def stop(self):
        logger.debug('Stop mode')
        resp = await self.send_command(Command.STOP_CURRENT_MODE)
        self._check_success(resp)

    async def set_color(self, mode: ColorTarget, r, g, b, brightness):
        # scale_light = [0x00, 0x32, 0x64]
        scale_light = [0x64, 0x64, 0x64]

        rgb_start = rgb_mid = rgb_end = \
            (b << 24) + (g << 16) + (r << 8) + brightness

        resp = await self.send_command(
            Command.WRITE_COLOR,
            struct.pack(
                '<BBIBIBI',
                mode.value,
                scale_light[0],
                rgb_start,
                scale_light[1],
                rgb_mid,
                scale_light[2],
                rgb_end,
            ),
        )
        self._check_zero_response(resp, 'Cannot set color')

    async def get_statistics(self):
        logger.debug('Get statistics')
        # b'\x00\x00\xdf\x01\x00\x00$\x01\x00\x00\t\x00\x00\x00\x00\x00'
        resp = await self.send_command(Command.GET_STATISTICS, b'\0')
        _, seconds_run, watts_hours, starts, _, _ = \
            struct.unpack('<HIIHHH', resp)
        return {
            'watts_hours': watts_hours,
            'seconds_run': seconds_run,
            'starts': starts,
        }

    async def get_starts_count(self):
        logger.debug('Get number of start')
        # b'\x00\x00\x00\t\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        # b'\x00\x00\x00\n\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        resp = await self.send_command(Command.GET_STARTS_COUNT, b'\0')
        _, _, starts, *_ = struct.unpack('<BHHHHHHHB', resp)
        return starts



import asyncio as aio
import logging
import uuid

from ..devices.base import BaseDevice
from ..devices.uuids import DEVICE_NAME

logger = logging.getLogger(__name__)

FIRMWARE_VERSION = uuid.UUID('00002a26-0000-1000-8000-00805f9b34fb')


class XiaomiPoller(BaseDevice):
    DATA_CHAR: uuid.UUID = None
    BATTERY_CHAR: uuid.UUID = None
    RECONNECTION_TIMEOUT = 60
    MANUFACTURER = 'Xiaomi'

    def __init__(self, *args, loop, **kwargs):
        super().__init__(*args, loop=loop, **kwargs)
        self._stack = aio.LifoQueue(loop=loop)
        self.connection_event = aio.Event()

    async def get_device_data(self):
        await self.client.start_notify(
            self.DATA_CHAR,
            self.notification_handler,
        )
        version = await self._read_with_timeout(FIRMWARE_VERSION)
        if isinstance(version, (bytes, bytearray)):
            self._version = version.decode()
        name = await self._read_with_timeout(DEVICE_NAME)
        if isinstance(name, (bytes, bytearray)):
            self._model = name.decode()

    def filter_notifications(self, sender):
        return True

    def notification_handler(self, sender, data: bytearray):
        logger.debug("Notification: {0}: {1}".format(
            sender,
            ' '.join(format(x, '02x') for x in data),
        ))
        if self.filter_notifications(sender):
            self._stack.put_nowait(data)

    async def read_and_send_data(self, publish_topic):
        raise NotImplementedError()

    async def handle(self, publish_topic, *args, **kwargs):
        while True:
            try:
                logger.debug(f'Wait {self} for connecting...')
                await aio.wait_for(
                    self.connection_event.wait(),
                    timeout=1,
                )
            except aio.TimeoutError:
                continue

            try:
                logger.debug(f'{self} connected!')
                await self.read_and_send_data(publish_topic)
            except ValueError as e:
                logger.error(f'Cannot read values {str(e)}')
            else:
                if await self.connection_event.wait():
                    await self.close()
                    return
            await aio.sleep(1)



from .base import registered_device_types  # noqa
from .kettle_redmond import RedmondKettle  # noqa
from .xiaomi_ht import XiaomiHumidityTemperatureV1  # noqa
from .xiaomi_lywsd03 import XiaomiHumidityTemperatureLYWSD  # noqa



import logging
import struct
import uuid
from dataclasses import dataclass

from ..utils import cr2032_voltage_to_percent
from .base import Device
from .xiaomi_base import XiaomiHumidityTemperature

logger = logging.getLogger(__name__)

LYWSD_DATA = uuid.UUID('EBE0CCC1-7A0A-4B0C-8A1A-6FF2997DA3A6')
LYWSD_BATTERY = uuid.UUID('EBE0CCC4-7A0A-4B0C-8A1A-6FF2997DA3A6')


@dataclass
class SensorState:
    battery: int
    temperature: float
    humidity: float

    @classmethod
    def from_data(cls, sensor_data, battery_data):
        t, h, voltage = struct.unpack('<hBH', sensor_data)
        return cls(
            temperature=round(t/100, 2),
            humidity=h,
            battery=int(cr2032_voltage_to_percent(voltage)),
        )


class XiaomiHumidityTemperatureLYWSD(XiaomiHumidityTemperature, Device):
    NAME = 'xiaomilywsd'
    REQUIRE_CONNECTION = True
    DATA_CHAR = LYWSD_DATA
    BATTERY_CHAR = LYWSD_BATTERY
    SENSOR_CLASS = SensorState
